@@ -187,7 +187,7 @@ tp_account_channel_request_get_property (GObject *object,
         break;
 
       case PROP_REQUEST:
-        g_value_set_object (value, self->priv->request);
+        g_value_set_boxed (value, self->priv->request);
         break;
 
       case PROP_USER_ACTION_TIME:
@@ -522,13 +522,15 @@ handle_channels (TpSimpleHandler *handler,
       return;
     }
 
+  tp_handle_channels_context_accept (context);
+
   if (self->priv->result == NULL)
     {
       /* We are re-handling the channel, no async request to complete */
       g_signal_emit (self, signals[SIGNAL_RE_HANDLED], 0, self->priv->channel,
           user_action_time, context);
 
-      goto out;
+      return;
     }
 
   /* Request succeeded */
@@ -545,9 +547,6 @@ handle_channels (TpSimpleHandler *handler,
     }
 
   handle_request_complete (self, channel, context);
-
-out:
-  tp_handle_channels_context_accept (context);
 }
 
 static void
@@ -825,6 +824,9 @@ request_and_handle_channel_finish (TpAccountChannelRequest *self,
  * acting like tp_account_channel_request_create_channel_async() with the
  * temporary #TpBaseClient as the @preferred_handler.)
  *
+ * The caller is responsible for closing the channel with
+ * tp_cli_channel_call_close() when it has finished handling it.
+ *
  * Since: 0.11.12
  */
 void
@@ -852,6 +854,9 @@ tp_account_channel_request_create_and_handle_channel_async (
  * See tp_account_channel_request_ensure_and_handle_channel_finish()
  * for details of how @context can be used.
  *
+ * The caller is responsible for closing the channel with
+ * tp_cli_channel_call_close() when it has finished handling it.
+ *
  * Returns: (transfer full) (allow-none): a new reference on a #TpChannel if the
  * channel was successfully created and you are handling it, otherwise %NULL.
  *
