@@ -160,6 +160,7 @@
 #include "telepathy-glib/base-client-internal.h"
 
 #include <stdarg.h>
+#include <string.h>                     /* for G_VA_COPY */
 
 #include <dbus/dbus.h>
 #include <dbus/dbus-glib-lowlevel.h>
@@ -891,7 +892,7 @@ tp_base_client_init (TpBaseClient *self)
   g_ptr_array_add (self->priv->handler_caps, NULL);
 
   self->priv->my_chans = g_hash_table_new_full (g_str_hash, g_str_equal,
-      NULL, NULL);
+      NULL, g_object_unref);
 }
 
 static void
@@ -911,6 +912,10 @@ tp_base_client_dispose (GObject *object)
   g_list_free (self->priv->pending_requests);
   self->priv->pending_requests = NULL;
 
+  if (self->priv->my_chans != NULL &&
+      g_hash_table_size (self->priv->my_chans) > 0)
+    WARNING ("TpBaseClient is still handling channels at dispose");
+
   tp_clear_pointer (&self->priv->my_chans, g_hash_table_unref);
 
   if (dispose != NULL)
@@ -934,6 +939,10 @@ tp_base_client_finalize (GObject *object)
   g_free (self->priv->bus_name);
   g_free (self->priv->object_path);
 
+  tp_clear_pointer (&self->priv->account_features, g_array_unref);
+  tp_clear_pointer (&self->priv->connection_features, g_array_unref);
+  tp_clear_pointer (&self->priv->channel_features, g_array_unref);
+
   if (finalize != NULL)
     finalize (object);
 }
@@ -1473,7 +1482,10 @@ _tp_base_client_observe_channels (TpSvcClientObserver *iface,
   connection = tp_account_ensure_connection (account, connection_path);
   if (connection == NULL)
     {
-      DEBUG ("Failed to create TpConnection");
+      g_set_error (&error, TP_ERRORS, TP_ERROR_INVALID_ARGUMENT,
+          "Connection %s doesn't seem to exist. (Maybe the CM doesn't own "
+          "the corresponding bus name?)", connection_path);
+      DEBUG ("Failed to create TpConnection: %s", error->message);
       goto out;
     }
 
@@ -2379,24 +2391,23 @@ varargs_helper (TpBaseClient *self,
 {
   GQuark f;
   gsize n = 0;
+  GQuark *features;
   va_list ap_copy;
 
-  va_copy (ap_copy, ap);
+  G_VA_COPY (ap_copy, ap);
 
   for (f = feature; f != 0; f = va_arg (ap, GQuark))
     n++;
 
-  /* block to provide a scope for @features on the stack */
-    {
-      GQuark features[n];
+  features = g_malloc_n (n, sizeof (GQuark));
 
-      n = 0;
+  n = 0;
 
-      for (f = feature; f != 0; f = va_arg (ap_copy, GQuark))
-        features[n++] = f;
+  for (f = feature; f != 0; f = va_arg (ap_copy, GQuark))
+    features[n++] = f;
 
-      method (self, features, n);
-    }
+  method (self, features, n);
+  g_free (features);
 }
 
 /**
